package flink.zanxiangnet.ad.monitoring;

import com.aliyun.odps.Instance;
import com.aliyun.odps.Odps;
import com.aliyun.odps.account.Account;
import com.aliyun.odps.account.AliyunAccount;
import com.aliyun.odps.data.Record;
import com.aliyun.odps.task.SQLTask;
import com.tencent.ads.model.DailyReportsGetListStruct;
import com.tencent.ads.model.HourlyReportsGetListStruct;
import flink.zanxiangnet.ad.monitoring.maxcompute.MaxComputeLog;
import flink.zanxiangnet.ad.monitoring.pojo.dto.AdDataOfDayDTO;
import flink.zanxiangnet.ad.monitoring.pojo.dto.AdStatOfDayODSDTO;
import flink.zanxiangnet.ad.monitoring.pojo.entity.*;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import flink.zanxiangnet.ad.monitoring.pojo.properties.KafkaProperties;
import flink.zanxiangnet.ad.monitoring.pojo.dto.AdDataOfMinuteDTO;
import flink.zanxiangnet.ad.monitoring.process.AdMinuteDWDProcess;
import flink.zanxiangnet.ad.monitoring.sink.TunnelBatchStreamSink;
import flink.zanxiangnet.ad.monitoring.stream.KeyedBatchStream;
import flink.zanxiangnet.ad.monitoring.util.DateUtil;
import flink.zanxiangnet.ad.monitoring.util.JsonUtil;
import org.apache.commons.lang3.StringUtils;
import org.apache.flink.api.common.eventtime.*;
import org.apache.flink.api.common.serialization.SimpleStringSchema;
import org.apache.flink.api.common.state.MapState;
import org.apache.flink.api.common.state.MapStateDescriptor;
import org.apache.flink.api.common.state.ValueState;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.api.common.typeinfo.Types;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.connector.kafka.source.KafkaSource;
import org.apache.flink.connector.kafka.source.enumerator.initializer.OffsetsInitializer;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.datastream.DataStreamSource;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.functions.ProcessFunction;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.assigners.TumblingEventTimeWindows;
import org.apache.flink.streaming.api.windowing.time.Time;
import org.apache.flink.streaming.api.windowing.triggers.Trigger;
import org.apache.flink.streaming.api.windowing.triggers.TriggerResult;
import org.apache.flink.streaming.api.windowing.windows.GlobalWindow;
import org.apache.flink.streaming.api.windowing.windows.TimeWindow;
import org.apache.flink.util.Collector;
import org.apache.flink.util.OutputTag;
import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.consumer.OffsetResetStrategy;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.common.config.SaslConfigs;
import org.apache.kafka.common.config.SslConfigs;
import org.springframework.beans.BeanUtils;

import java.time.Duration;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.*;
import java.util.stream.Collectors;

public class AdStatJob {

    /**
     * 可能有数据的最早日期
     */
    private static final String OLDEST_DAY = "2019-01-01";

    public static void main(String[] args) throws Exception {
        System.setProperty("javax.net.ssl.trustStore", "D:\\Downloads\\kafka.client.truststore.jks");
        // System.setProperty("javax.net.ssl.trustStore", "/root/flink-1.13.2/kafka.client.truststore.jks");
        System.setProperty("javax.net.ssl.trustStorePassword", "KafkaOnsClient");
        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();

        // 加载配置文件到 flink的全局配置中
        Properties props = new Properties();
        props.load(AdStatJob.class.getResourceAsStream("/application.properties"));
        Configuration configuration = new Configuration();
        props.stringPropertyNames().forEach(key -> {
            String value = props.getProperty(key);
            configuration.setString(key.trim(), StringUtils.isBlank(value) ? "" : value.trim());
        });
        env.getConfig().setGlobalJobParameters(configuration);

        Properties adStreamOfMinuteProps = new Properties();
        adStreamOfMinuteProps.load(AdStatJob.class.getResourceAsStream("/ad_stream_of_minute.properties"));
        KafkaSource<String> adStreamOfMinuteSource = buildKafkaSource(adStreamOfMinuteProps);

        DataStreamSource<String> adStreamOfMinuteIn = env.fromSource(adStreamOfMinuteSource, WatermarkStrategy.noWatermarks(), "adStreamOfMinuteSource_kafka");

        // 广告分钟数据（前 5分钟的广告消耗数据）
        final OutputTag<AdDataOfMinuteODS> adMinuteStreamTag = new OutputTag<AdDataOfMinuteODS>("adMinuteStream") {
        };
        // 广告小时数据（往前回滚 10天）
        final OutputTag<AdDataOfHourODS> adHourStreamTag = new OutputTag<AdDataOfHourODS>("adHourStream") {
        };

        // 对流进行映射，拆分（实时的分钟流和回滚的小时流）
        SingleOutputStreamOperator<AdDataOfMinuteODS> adODSStream = adStreamOfMinuteIn.filter(StringUtils::isNotBlank)
                .process(new ProcessFunction<String, AdDataOfMinuteODS>() {
                    @Override
                    public void processElement(String adDataStr, ProcessFunction<String, AdDataOfMinuteODS>.Context context, Collector<AdDataOfMinuteODS> collector) throws Exception {
                        AdDataOfMinuteDTO dto = JsonUtil.toObj(adDataStr, AdDataOfMinuteDTO.class);
                        // 指记录被创建的时间
                        long createTime = dto.getCreateTime();
                        // 指记录统计的时间（如果是实时统计的数据精确到分钟，回滚的历史数据精确到天）
                        long statTime = DateUtil.localDateTimeToMilli(DateUtil.milliToLocalDateTime(dto.getDataTime()).withSecond(0).withNano(0));
                        HourlyReportsGetListStruct struct = dto.getHourlyReportsGetListStruct();

                        if (createTime - statTime > (60 * 60 * 1000L)) {
                            AdDataOfHourODS adODS = new AdDataOfHourODS();
                            BeanUtils.copyProperties(struct, adODS);
                            adODS.setStatDay(DateUtil.formatLocalDate(DateUtil.milliToLocalDate(statTime)));
                            adODS.setHour(dto.getHourlyReportsGetListStruct().getHour().intValue());
                            adODS.setStatTime(DateUtil.localDateTimeToMilli(DateUtil.milliToLocalDateTime(statTime).withMinute(0)));
                            adODS.setAccountId(dto.getAccountId());
                            adODS.setAgencyAccountId(dto.getHourlyReportsGetListStruct().getAccountId());
                            adODS.setCreateTime(new Date(createTime));
                            context.output(adHourStreamTag, adODS);
                        } else {
                            AdDataOfMinuteODS adODS = new AdDataOfMinuteODS();
                            BeanUtils.copyProperties(struct, adODS);
                            adODS.setStatDay(DateUtil.formatLocalDate(DateUtil.milliToLocalDate(statTime)));
                            adODS.setHour(dto.getHourlyReportsGetListStruct().getHour().intValue());
                            adODS.setStatTime(statTime);
                            adODS.setAccountId(dto.getAccountId());
                            adODS.setAgencyAccountId(dto.getHourlyReportsGetListStruct().getAccountId());
                            adODS.setCreateTime(new Date(createTime));
                            context.output(adMinuteStreamTag, adODS);
                        }
                    }
                });

        // 分钟流
        DataStream<AdDataOfMinuteODS> adMinuteODSStream = adODSStream.getSideOutput(adMinuteStreamTag);
        // 写入原始表
        // adMinuteODSStream.addSink(new TunnelBatchSink<>(AdDataOfMinuteODS.class, 36000L, 64000L, 3));
        new KeyedBatchStream<>("adMinuteODSStream", adMinuteODSStream.keyBy(AdDataOfMinuteODS::getStatDay), 4000L, 2 * 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdDataOfMinuteODS.class));

        SingleOutputStreamOperator<AdStatOfMinuteDWD> adMinuteDWDStream = adMinuteODSStream
                // 打水印，允许数据延迟 6分钟，同时指定时间流
                .assignTimestampsAndWatermarks(WatermarkStrategy.<AdDataOfMinuteODS>forBoundedOutOfOrderness(Duration.ofMinutes(1L))
                        .withTimestampAssigner((SerializableTimestampAssigner<AdDataOfMinuteODS>) (adODS, l) -> adODS.getStatTime()))
                .keyBy(AdDataOfMinuteODS::getAdId)
                // 开一个 5分钟的滚动窗口
                .window(TumblingEventTimeWindows.of(Time.minutes(5L)))
                .trigger(new Trigger<AdDataOfMinuteODS, TimeWindow>() {
                    @Override
                    public TriggerResult onElement(AdDataOfMinuteODS adDataOfMinuteODS, long time, TimeWindow timeWindow, TriggerContext triggerContext) throws Exception {
                        // FIXME 水印时间没有意义！拿到数据是 Long.MAX
                        if (timeWindow.maxTimestamp() <= triggerContext.getCurrentWatermark()) {
                            // 到了窗口的最大生命周期
                            return TriggerResult.FIRE_AND_PURGE;
                        }
                        triggerContext.registerEventTimeTimer(timeWindow.maxTimestamp());
                        if (adDataOfMinuteODS.getHour() == DateUtil.milliToLocalDateTime(time).getHour()) {
                            return TriggerResult.FIRE;
                        }
                        return TriggerResult.CONTINUE;
                    }

                    @Override
                    public TriggerResult onProcessingTime(long time, TimeWindow timeWindow, TriggerContext triggerContext) throws Exception {
                        return TriggerResult.CONTINUE;
                    }

                    @Override
                    public TriggerResult onEventTime(long time, TimeWindow timeWindow, TriggerContext triggerContext) throws Exception {
                        return time == timeWindow.maxTimestamp() ? TriggerResult.FIRE : TriggerResult.CONTINUE;
                    }

                    @Override
                    public void clear(TimeWindow timeWindow, TriggerContext triggerContext) throws Exception {
                        triggerContext.deleteEventTimeTimer(timeWindow.maxTimestamp());
                    }
                })
                .process(new AdMinuteDWDProcess());
        // .addSink(new TunnelBatchSink<>(AdStatOfMinuteDWD.class, 30000L, 365L, 6));
        new KeyedBatchStream<>("adMinuteDWDStream", adMinuteDWDStream.keyBy(AdStatOfMinuteDWD::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdStatOfMinuteDWD.class));

        // 小时流（直接写到小时报表的 ods）
        DataStream<AdDataOfHourODS> adHourODSStream = adODSStream.getSideOutput(adHourStreamTag);
        // 写入原始表
        // adHourODSStream.addSink(new TunnelBatchSink<>(AdDataOfHourODS.class, 36000L, 64000L, 10));
        new KeyedBatchStream<>("adHourODSStream", adHourODSStream.keyBy(AdDataOfHourODS::getStatDay), 4000L, 5 * 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdDataOfHourODS.class));

        SingleOutputStreamOperator<AdStatOfHourDWD> adHourDWDStream = adHourODSStream.keyBy(AdDataOfHourODS::getAdId)
                .countWindow(1).process(new ProcessWindowFunction<AdDataOfHourODS, AdStatOfHourDWD, Long, GlobalWindow>() {
                    private Odps odps;
                    // 上次查询的天数据
                    private ValueState<String> lastQueryDayState;
                    // 聚合的天的数据
                    private MapState<String, AdStatOfHourDWD> historyReduceState;

                    @Override
                    public void open(Configuration conf) {
                        Map<String, String> params = getRuntimeContext()
                                .getExecutionConfig()
                                .getGlobalJobParameters()
                                .toMap();
                        Account account = new AliyunAccount(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ID),
                                params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_KEY));
                        odps = new Odps(account);
                        odps.getRestClient().setRetryLogger(new MaxComputeLog());
                        odps.setEndpoint(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ENDPOINT));
                        odps.setDefaultProject(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_PROJECT_NAME));

                        lastQueryDayState = getRuntimeContext().getState(new ValueStateDescriptor<>("lastQueryDayState", String.class));
                        historyReduceState = getRuntimeContext().getMapState(new MapStateDescriptor<>("historyReduceState", Types.STRING, Types.POJO(AdStatOfHourDWD.class)));
                    }

                    @Override
                    public void process(Long elementsCount, ProcessWindowFunction<AdDataOfHourODS, AdStatOfHourDWD, Long, GlobalWindow>.Context context, Iterable<AdDataOfHourODS> iterable, Collector<AdStatOfHourDWD> collector) throws Exception {
                        AdDataOfHourODS element = iterable.iterator().next();
                        LocalDate statDay = DateUtil.parseLocalDate(element.getStatDay());
                        LocalDateTime statTime = LocalDateTime.of(statDay, LocalTime.of(element.getHour(), 0, 0));
                        long now = System.currentTimeMillis();
                        LocalDate today = LocalDate.now();

                        String lastQueryDay = lastQueryDayState.value();
                        // 从 maxCompute拉取指定 广告的历史数据
                        if (lastQueryDay == null || !lastQueryDay.equals(DateUtil.formatLocalDate(today))) {
                            LocalDate endDay = today, beginDay = statDay.minusDays(60);
                            Instance instance = SQLTask.run(odps, "SELECT * FROM ad_stat_of_hour_dwd WHERE stat_day >= \"" + DateUtil.formatLocalDate(beginDay) + "\" AND stat_day <= \"" + DateUtil.formatLocalDate(endDay) + "\" AND ad_id = " + element.getAdId() + ";");
                            instance.waitForSuccess();
                            List<Record> records = SQLTask.getResult(instance);
                            Map<String, AdStatOfHourDWD> historyHourMap = records.stream()
                                    .map(AdStatOfHourDWD::byMaxCompute)
                                    .sorted((o1, o2) -> new Long(o1.getCreateTime().getTime() - o2.getCreateTime().getTime()).intValue())
                                    .collect(Collectors.toMap(data -> data.getStatDay() + data.getHour(), data -> data, (val1, val2) -> val2));
                            historyReduceState.clear();
                            historyReduceState.putAll(historyHourMap);
                            lastQueryDayState.update(DateUtil.formatLocalDate(today));
                        }
                        AdStatOfHourDWD lastReduceData = null;
                        for (int i = 1; i < 60 * 24; i++) {
                            LocalDateTime time = statTime.minusHours(i);
                            lastReduceData = historyReduceState.get(DateUtil.formatLocalDate(time.toLocalDate()) + time.getHour());
                            if (lastReduceData != null) {
                                break;
                            }
                        }

                        AdStatOfHourDWD newStatData = AdStatOfHourDWD.reduce(lastReduceData, element, now);
                        collector.collect(newStatData);
                    }
                });
        //.addSink(new TunnelBatchSink<>(AdStatOfHourDWD.class, 30000L, 365L, 6));
        new KeyedBatchStream<>("adHourDWDStream", adHourDWDStream.keyBy(AdStatOfHourDWD::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdStatOfHourDWD.class));


        // ------------------------------------------------------- 处理广告的天数据 -----------------------------------------
        Properties adStreamOfDayProps = new Properties();
        adStreamOfDayProps.load(AdStatJob.class.getResourceAsStream("/ad_stream_of_day.properties"));
        KafkaSource<String> adStreamOfDaySource = buildKafkaSource(adStreamOfDayProps);

        DataStreamSource<String> adStreamOfDayIn = env.fromSource(adStreamOfDaySource, WatermarkStrategy.noWatermarks(), "adStreamOfMinuteSource_kafka");

        // 广告日数据。往前回滚 10天
        final OutputTag<AdDataOfDayODS> adDayStreamRollDayTag = new OutputTag<AdDataOfDayODS>("adDayStreamRollDayTag") {
        };
        // 广告日数据。往前回滚 1年
        final OutputTag<AdDataOfDayODS> adDayStreamRollYearTag = new OutputTag<AdDataOfDayODS>("adDayStreamRollYearTag") {
        };
        SingleOutputStreamOperator<AdStatOfDayODSDTO> adDayODSStream = adStreamOfDayIn.filter(StringUtils::isNotBlank)
                .map(str -> {
                    AdDataOfDayDTO dto = JsonUtil.toObj(str, AdDataOfDayDTO.class);
                    Date createTime = new Date(dto.getCreateTime());
                    DailyReportsGetListStruct struct = dto.getDailyReportsGetListStruct();
                    AdDataOfDayODS adODS = new AdDataOfDayODS();
                    BeanUtils.copyProperties(struct, adODS);
                    adODS.setStatDay(struct.getDate());
                    adODS.setAccountId(dto.getAccountId());
                    adODS.setCampaignId(struct.getCampaignId());
                    adODS.setAgencyAccountId(struct.getAccountId());
                    adODS.setWechatAccountId(struct.getWechatAccountId());
                    adODS.setAdgroupId(struct.getAdgroupId());
                    adODS.setAdId(struct.getAdId());
                    adODS.setCreateTime(createTime);
                    return AdStatOfDayODSDTO.builder()
                            .startDate(dto.getStartDate())
                            .endDate(dto.getEndDate())
                            .adDataOfDayODS(adODS)
                            .build();
                });
        // adDayODSStream.map(AdStatOfDayODSDTO::getAdDataOfDayODS).addSink(new TunnelBatchSink<>(AdDataOfDayODS.class, 36000L, 6000L, 10));
        new KeyedBatchStream<>("adDayODSStream", adDayODSStream.map(AdStatOfDayODSDTO::getAdDataOfDayODS).keyBy(AdDataOfDayODS::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdDataOfDayODS.class));

        // 拆分流
        SingleOutputStreamOperator<AdStatOfDayODSDTO> adDayODSStreamSplit = adDayODSStream.process(new ProcessFunction<AdStatOfDayODSDTO, AdStatOfDayODSDTO>() {
            @Override
            public void processElement(AdStatOfDayODSDTO adDataDTO, ProcessFunction<AdStatOfDayODSDTO, AdStatOfDayODSDTO>.Context context, Collector<AdStatOfDayODSDTO> collector) throws Exception {
                LocalDate startDate = adDataDTO.getStartDate();
                LocalDate endDate = adDataDTO.getEndDate();
                AdDataOfDayODS adODS = adDataDTO.getAdDataOfDayODS();
                if (DateUtil.intervalOfDays(startDate, endDate) > 31L) {
                    // 拉取时间间隔超过 1个月，账号回滚 365天的数据
                    context.output(adDayStreamRollYearTag, adODS);
                } else {
                    // 每日往前回滚 10天的数据
                    context.output(adDayStreamRollDayTag, adODS);
                }
            }
        });

        SingleOutputStreamOperator<AdStatOfDayDWD> adDayDWDStream = adDayODSStreamSplit.getSideOutput(adDayStreamRollDayTag)
                .keyBy(AdDataOfDayODS::getAdId)
                .countWindow(1).process(new ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow>() {
                    private Odps odps;
                    // 上次查询的天数据
                    private ValueState<String> lastQueryDayState;
                    // 之前聚合的昨天的数据
                    private MapState<String, AdStatOfDayDWD> historyReduceState;

                    @Override
                    public void open(Configuration conf) {
                        Map<String, String> params = getRuntimeContext()
                                .getExecutionConfig()
                                .getGlobalJobParameters()
                                .toMap();
                        Account account = new AliyunAccount(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ID),
                                params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_KEY));
                        odps = new Odps(account);
                        odps.getRestClient().setRetryLogger(new MaxComputeLog());
                        odps.setEndpoint(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ENDPOINT));
                        odps.setDefaultProject(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_PROJECT_NAME));

                        lastQueryDayState = getRuntimeContext().getState(new ValueStateDescriptor<>("lastQueryDayState", String.class));
                        historyReduceState = getRuntimeContext().getMapState(new MapStateDescriptor<>("historyReduceState", String.class, AdStatOfDayDWD.class));
                    }

                    @Override
                    public void process(Long elementsCount, ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow>.Context context, Iterable<AdDataOfDayODS> iterable, Collector<AdStatOfDayDWD> collector) throws Exception {
                        AdDataOfDayODS element = iterable.iterator().next();
                        LocalDate statDay = DateUtil.parseLocalDate(element.getStatDay());
                        long now = System.currentTimeMillis();

                        String lastQueryDay = lastQueryDayState.value();
                        // 从 maxCompute查找广告的历史数据
                        if (lastQueryDay == null || !lastQueryDay.equals(DateUtil.formatLocalDate(LocalDate.now()))) {
                            LocalDate endTime = LocalDate.now(), beginTime = statDay.minusDays(60);
                            Instance instance = SQLTask.run(odps, "SELECT * FROM ad_stat_of_day_dwd WHERE stat_day >= \"" + DateUtil.formatLocalDate(beginTime) + "\" AND stat_day <= \"" + DateUtil.formatLocalDate(endTime) + "\" AND ad_id = " + element.getAdId() + ";");
                            instance.waitForSuccess();
                            List<Record> records = SQLTask.getResult(instance);
                            Map<String, AdStatOfDayDWD> historyData = records.stream()
                                    .map(AdStatOfDayDWD::byMaxCompute)
                                    .sorted((o1, o2) -> new Long(o1.getCreateTime().getTime() - o2.getCreateTime().getTime()).intValue())
                                    .collect(Collectors.toMap(AdStatOfDayDWD::getStatDay, data -> data, (val1, val2) -> val2));
                            historyReduceState.clear();
                            historyReduceState.putAll(historyData);
                            lastQueryDayState.update(DateUtil.formatLocalDate(LocalDate.now()));
                        }
                        AdStatOfDayDWD lastReduceData = null;
                        for (int i = 1; i <= 60; i++) {
                            lastReduceData = historyReduceState.get(DateUtil.formatLocalDate(statDay.minusDays(i)));
                            if (lastReduceData != null) {
                                break;
                            }
                        }
                        AdStatOfDayDWD newStatData = AdStatOfDayDWD.reduce(lastReduceData, element, now);
                        historyReduceState.put(DateUtil.formatLocalDate(statDay), newStatData);
                        collector.collect(newStatData);
                    }
                });
        //.addSink(new TunnelBatchSink<>(AdStatOfDayDWD.class, 30000L, 365L, 6));
        new KeyedBatchStream<>("adDayDWDStream", adDayDWDStream.keyBy(AdStatOfDayDWD::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdStatOfDayDWD.class));

        SingleOutputStreamOperator<AdStatOfDayDWD> adDayDWDYearStream = adDayODSStreamSplit.getSideOutput(adDayStreamRollYearTag)
                .keyBy(AdDataOfDayODS::getAdId)
                .countWindow(1).process(new ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow>() {
                    // 上次聚合的结果
                    private AdStatOfDayDWD lastReduce;

                    @Override
                    public void process(Long elementsCount, ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow>.Context context, Iterable<AdDataOfDayODS> elements, Collector<AdStatOfDayDWD> out) {
                        AdDataOfDayODS element = elements.iterator().next();
                        lastReduce = AdStatOfDayDWD.reduce(lastReduce, element, System.currentTimeMillis());
                        out.collect(lastReduce);
                    }
                });
        //.addSink(new TunnelBatchSink<>(AdStatOfDayDWD.class, 30000L, 365L, 6));
        new KeyedBatchStream<>("adDayDWDYearStream", adDayDWDYearStream.keyBy(AdStatOfDayDWD::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdStatOfDayDWD.class));

        env.execute();
    }

    private static KafkaSource<String> buildKafkaSource(Properties props) {
        return buildKafkaSource(props.getProperty(KafkaProperties.KAFKA_SERVERS),
                props.getProperty(KafkaProperties.KAFKA_USERNAME),
                props.getProperty(KafkaProperties.KAFKA_PASSWORD),
                props.getProperty(KafkaProperties.KAFKA_SSL_PATH),
                props.getProperty(KafkaProperties.KAFKA_TOPIC),
                props.getProperty(KafkaProperties.KAFKA_GROUP_ID)
        );
    }

    /**
     * 内网
     *
     * @param servers
     * @param username
     * @param password
     * @param sslPath
     * @param topic
     * @param groupId
     * @return
     */
    private static KafkaSource<String> buildKafkaSourceVPC(String servers, String username, String password, String sslPath, String topic, String groupId) {
        Properties props = new Properties();
        props.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, servers);
        //两次Poll之间的最大允许间隔。
        //消费者超过该值没有返回心跳，服务端判断消费者处于非存活状态，服务端将消费者从Group移除并触发Rebalance，默认30s。
        props.put(ConsumerConfig.SESSION_TIMEOUT_MS_CONFIG, 30000);
        //每次Poll的最大数量。
        //注意该值不要改得太大，如果Poll太多数据，而不能在下次Poll之前消费完，则会触发一次负载均衡，产生卡顿。
        props.put(ConsumerConfig.MAX_POLL_RECORDS_CONFIG, 30);
        //消息的反序列化方式。
        props.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringDeserializer");
        props.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringDeserializer");
        //当前消费实例所属的消费组，请在控制台申请之后填写。
        //属于同一个组的消费实例，会负载消费消息。
        props.put(ConsumerConfig.GROUP_ID_CONFIG, groupId);
        return KafkaSource.<String>builder()
                .setProperties(props)
                .setBootstrapServers(servers)
                .setTopics(topic)
                .setGroupId(groupId)
                .setStartingOffsets(OffsetsInitializer.committedOffsets(OffsetResetStrategy.EARLIEST))
                .setValueOnlyDeserializer(new SimpleStringSchema())
                .build();
    }

    /**
     * 公网
     *
     * @param servers
     * @param username
     * @param password
     * @param sslPath
     * @param topic
     * @param groupId
     * @return
     */
    private static KafkaSource<String> buildKafkaSource(String servers, String username, String password, String sslPath, String topic, String groupId) {
        Properties props = new Properties();
        props.put(SaslConfigs.SASL_JAAS_CONFIG, "org.apache.kafka.common.security.plain.PlainLoginModule required username=\"" + username + "\" password=\"" + password + "\";");
        props.put(SslConfigs.SSL_TRUSTSTORE_LOCATION_CONFIG, sslPath);
        props.put(SslConfigs.SSL_TRUSTSTORE_PASSWORD_CONFIG, "KafkaOnsClient");
        props.put(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, "SASL_SSL");
        props.put(SaslConfigs.SASL_MECHANISM, "PLAIN");
        props.put(SslConfigs.SSL_ENDPOINT_IDENTIFICATION_ALGORITHM_CONFIG, "");
        props.put(ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringSerializer");
        props.put(ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringSerializer");
        return KafkaSource.<String>builder()
                .setBootstrapServers(servers)
                .setTopics(topic)
                .setGroupId(groupId)
                .setProperties(props)
                .setStartingOffsets(OffsetsInitializer.committedOffsets(OffsetResetStrategy.EARLIEST))
                .setValueOnlyDeserializer(new SimpleStringSchema())
                .build();
    }
}
